﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using UnityEngine.Audio;

public class DSP : Singleton<DSP> {

    public AudioClip clip;

    float[] spectrumData;

    public float voiceAcknowledgementThreshold = 0.01f;
    Dictionary<float, float> frequencyAmplitudes;
    static float peakEnergyCutoffPercentile = 0.65f;
    public float newObservationWeight = 0.5f;
    int fftSize_Pow = 13;
    int fftSize { get { return (int)Mathf.Pow(2, fftSize_Pow); } }
    int numBins { get { return fftSize / 2; } }
    int samplingRate = 44100;
    int band { get { return samplingRate / 2; } }
    // float FR { get { return band / (float)numBins; } }
    float FR { get { return samplingRate / (float)fftSize; } }
    public Vector2 frequencyLimits = new Vector2(50, 500);
    int maxNumPeaks { get { return Mathf.CeilToInt(frequencyLimits.y / frequencyLimits.x) + 1; } }
    // Frequency (i) = i * FR = 100 => i = 100 / FR
    int numBinFq_Min { get { return FrequencyToBin(frequencyLimits.x); } }
    int numBinFq_Max { get { return FrequencyToBin(frequencyLimits.y); } }
    private Dictionary<Player, Coroutine> recordingCoroutines = new Dictionary<Player, Coroutine>();

    private Dictionary<Player, MicrophoneSource> playerMicrophoneSources = new Dictionary<Player, MicrophoneSource>();

    public int FrequencyToBin (float frequency)
    {
        return Mathf.FloorToInt(frequency / FR);
    }
    public float BinToFrequency (int bin)
    {
        return bin * FR;
    }

    public static bool initialized { get; private set; }
    public int micDuration = 1;

    protected override void DeInitialize()
    {
        base.DeInitialize();
        StopAllCoroutines();
    }
    // Use this for initialization
    protected override void Initialize()
    {
        Debug.Log(Microphone.devices.ToReadableString());
        List<string> names = new List<string>();
        foreach (string s in Microphone.devices)
            if (!s.ContainsInvariant("realtek") && !s.ContainsInvariant("built"))
                names.Add(s);

        if (names.Count == 0)
            names = new List<string>(Microphone.devices);

        foreach (Player player in The_Helper.EnumGetValues<Player>())
        {
            playerMicrophoneSources.Add(player, 
                new MicrophoneSource(names[((int) player) % names.Count], micDuration, samplingRate));
            recordingCoroutines.Add(player, null);
        }

        frequencyAmplitudes = new Dictionary<float, float>();

        for (int i = numBinFq_Min; i < numBinFq_Max; i++)
            frequencyAmplitudes.Add(BinToFrequency(i), 0);
        
        initialized = true;
    }

    public static void Record(Player player, 
        float preparationDuration, float recordDuration, Action<float, float> callback)
    {
        if (!initialized) return;
        if (!instance) return;
        instance._Record(player, preparationDuration, recordDuration, callback);
    }

    private void _Record(Player player, 
        float preparationDuration, float recordDuration, Action<float, float> callback)
    {
        if (recordingCoroutines[player] != null)
            StopCoroutine(recordingCoroutines[player]);
        this.Log("Asked to record Player " + player, LogType.Log, recordDuration);
        recordingCoroutines[player] = StartCoroutine(
            __Record(playerMicrophoneSources[player], 
            preparationDuration, recordDuration, callback));
    }

    public IEnumerator __Record(MicrophoneSource microphone, 
        float preparationDuration, float recordDuration, Action<float, float> callback)
    {
        yield return new WaitForSeconds(preparationDuration);
        microphone.Start();
        while (Microphone.GetPosition(microphone.deviceName) <= 0)
            yield return null;

        this.Log("<b>" + microphone.deviceName + "</b> INITIALIZED !");

        spectrumData = new float[numBins];

        float duration = 0;
        float avgFrequency = 0;

        float currentVol = 0;

        for (float t = 0; t < recordDuration; t += Time.deltaTime)
        {
            // this.Log("Now recording.. " + (t / recordDuration).PercentileToPercent(), LogType.Log, Time.deltaTime * 2);
            float[] data = new float[1024];
            microphone.source.GetOutputData(data, 0);
            float volume = 0;
            foreach (float d in data)
                volume += d * d;
            volume /= data.Length;
            volume = Mathf.Sqrt(volume);

            currentVol = Mathf.Lerp(currentVol, volume, newObservationWeight);
            // Debug.Log(volume);
            if (currentVol < voiceAcknowledgementThreshold || microphone.deviceName.ContainsInvariant("real"))
            {
                yield return new WaitForEndOfFrame();
                continue;
            }
            microphone.source.GetSpectrumData(spectrumData, 0, FFTWindow.BlackmanHarris);

            float[] frequencyKeys = frequencyAmplitudes.Keys.ToArray();
            foreach (float frequencyKey in frequencyKeys)
                AddObservation(frequencyKey, 
                    spectrumData[FrequencyToBin(frequencyKey)]);

            // Fail - safe
            float normalizedFrequency = GetPeakFrequency();// NormalizeFrequency(GetPeakFrequency());
            if (normalizedFrequency >= 0)
            {
                this.Log(microphone.deviceName + " : " + normalizedFrequency,LogType.Log, Time.deltaTime * 1.5f);
                duration += Time.deltaTime;
                avgFrequency += Time.deltaTime * normalizedFrequency;
            }
            else
            {
                //this.Log("This shouldn't happen - silence should have been detected earlier.", LogType.Warning);
            }

            //this.Log(NormalizeFrequency(GetPeakFrequency()).PercentileToPercent("#"), LogType.Log, Time.deltaTime * 2);

            for (int i = 1; i < spectrumData.Length - 1; i++)
            {
                //Debug.DrawLine(new Vector3(i - 1, spectrumData[i] + 10, 0), new Vector3(i, spectrumData[i + 1] + 10, 0), Color.red);
                Debug.DrawLine(new Vector3(i - 1, Mathf.Log(spectrumData[i - 1]) + 10, 2), new Vector3(i, Mathf.Log(spectrumData[i]) + 10, 2), Color.cyan, Time.deltaTime * 1.5f);
                //Debug.DrawLine(new Vector3(Mathf.Log(i - 1), spectrumData[i - 1] - 10, 1), new Vector3(Mathf.Log(i), spectrumData[i] - 10, 1), Color.green);
                //Debug.DrawLine(new Vector3(Mathf.Log(i - 1), Mathf.Log(spectrumData[i - 1]), 3), new Vector3(Mathf.Log(i), Mathf.Log(spectrumData[i]), 3), Color.blue);
            }

            yield return new WaitForEndOfFrame();
        }

        if (duration > 0)
            avgFrequency /= duration;
        else
            avgFrequency = 0;

        this.Log("Finished recording.\nDuration: " + duration + " | Frequency: " + avgFrequency, LogType.Log, recordDuration);

        microphone.Stop();

        if (callback != null)
            callback(duration, avgFrequency);
    }

    public float NormalizeFrequency(float frequency)
    {
        this.Log("BASE FQ: " + frequency);
        return frequency.RetargetedTo_01(frequencyLimits);//.Smooth01(SmoothType.Log, 5);
    }

    public void AddObservation(float frequency, float newAmplitude)
    {
        frequencyAmplitudes[frequency] = Mathf.Lerp(
            frequencyAmplitudes[frequency],
            newAmplitude, newObservationWeight);
    }
    public float GetPeakFrequency()
    {
        // Get the peaks, these include harmonics
        Dictionary<float, float> peakCandidates = 
            frequencyAmplitudes.OrderByDescending(pair => pair.Value).Take(maxNumPeaks)
            .ToDictionary(pair => pair.Key, pair => pair.Value);

        // Remove any candidates that are less than a % of the max or under the silence threshold
        float[] keys = peakCandidates.Keys.ToArray();
        foreach (float key in keys)
            if (peakCandidates[key] < Mathf.Max(voiceAcknowledgementThreshold, 
                peakEnergyCutoffPercentile * peakCandidates.Values.ToArray()[0]))
                peakCandidates.Remove(key);

        // This is silence
        if (peakCandidates.Count == 0)
            return -1;

        // this.Log(peakCandidates.ToReadableString());
        // Track down the lowest frequency
        return peakCandidates.OrderBy(pair => pair.Key).Take(1)
            .ToDictionary(pair => pair.Key, pair => pair.Value).Keys.ToArray()[0];
    }
}

public class MicrophoneSource
{
    public MicrophoneSource (string deviceName, int micDuration, int samplingRate)
    {
        GameObject gO = new GameObject("Microphone Source (" + deviceName + ")");

        this.deviceName = deviceName;
        this.micDuration = micDuration;
        this.samplingRate = samplingRate;

        source = gO.AddComponent<AudioSource>();
        AudioMixer mixer = Resources.Load("Humm") as AudioMixer;
        source.outputAudioMixerGroup = mixer.FindMatchingGroups("Master")[0];

        source.loop = true;
    }

    public void Start()
    {
        this.Log("Starting up <b>" + deviceName + "</b>");
        source.clip = Microphone.Start(deviceName, true, micDuration, samplingRate);
        source.Play();
    }

    public void Stop()
    {
        Microphone.End(deviceName);
        source.Stop();
        this.Log("<b>" + deviceName + "</b> stopped.");
    }

    ~MicrophoneSource()
    {
        if (source)
            source.Stop();
        if (deviceName != "")
            Microphone.End(deviceName);
    }

    public string deviceName;
    public AudioSource source;
    public int micDuration;
    public int samplingRate;
}