﻿using System;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;

using Shout.Input;

using UniRx;
using UnityEngine.UI;

public class PlayerController : MonoBehaviour {
	public Action<Player> playerIsDeadCallback;

    private Sprite mouthOpen, mouthClosed, mouthErgh;

	public Player player;
	public float offsetAmount = 1f;

	public LaserController laserController;
	public PotionConveyorController potionConveyorController;
	public MicrophoneInput input;
	public Animator animator;

	private IDisposable disposable;
    private HealthManager healthManager;

    public SpriteRenderer neckRenderer, headRenderer;
    private AudioSource audioSource;
    private Transform outputHigh, outputMid, outputLow;
    private Transform instrumentHigh, instrumentMid, instrumentLow;
    public void Initialize(HealthManager healthManager,
        Transform outputHigh, Transform outputMid, Transform outputLow,
        Transform instrumentHigh, Transform instrumentMid, Transform instrumentLow,
        Sprite neck, Sprite mouthOpen, Sprite mouthClosed, Sprite mouthErgh)
	{

        InitializeInputStream();
	
		laserController.Initialize( player );

        this.healthManager = healthManager;

        neckRenderer.sprite = neck;
        this.mouthOpen = mouthOpen;
        this.mouthClosed = mouthClosed;
        this.mouthErgh = mouthErgh;

        this.outputHigh = outputHigh;
        this.outputMid  = outputMid;
        this.outputLow  = outputLow;

        this.instrumentHigh = instrumentHigh;
        this.instrumentMid  = instrumentMid;
        this.instrumentLow  = instrumentLow;

        if (player == Player.Right)
            transform.FindChild("Dam").localPosition += Vector3.right * 0.8f;

        audioSource = GetComponent<AudioSource>();
        ToggleMouth(MouthState.Idle);

        if (player == Player.Right)
            neckRenderer.transform.localPosition -= Vector3.forward * .05f;
        else
            neckRenderer.transform.localPosition += Vector3.forward * .05f;

        healthManager.maxHealth = 3;
        healthManager.health = 3;

		animator = this.GetComponent<Animator>();

	}

    private Color colorOpen = Color.white, colorClosed = new Color(.65f, .65f, .65f, .93f);

    public void ToggleMouth(MouthState state)
    {
        Debug.Log(player + " -> " + state);
        if (state == MouthState.Open)
        {
            headRenderer.sprite = mouthOpen;
            float interpolator = 0;
            float transitionDuration = .35f;
            Color currentColor = headRenderer.color;
            The_Helper.StartTimer(Time.fixedDeltaTime, a => {
                interpolator += Time.fixedDeltaTime / transitionDuration;
                headRenderer.color = Color.Lerp(currentColor, colorOpen, interpolator);
                neckRenderer.color = Color.Lerp(currentColor, colorOpen, interpolator);
            }, (int)(transitionDuration / Time.fixedDeltaTime));
        }
        else if (state == MouthState.Closed)
        {
            headRenderer.sprite = mouthClosed;
            audioSource.PlayOneShot(The_Helper.ResourceLoad<AudioClip>(
                "Release"+player), GetMouthSFXVolume(state));
        }
        else if (state == MouthState.Ergh)
        {
            headRenderer.sprite = mouthErgh;
            audioSource.PlayOneShot(The_Helper.ResourceLoad<AudioClip>(
                "Ergh" + player), GetMouthSFXVolume(state));
            float interpolator = 0;
            float transitionDuration = .35f;
            Color currentColor = headRenderer.color;
            The_Helper.StartTimer(Time.fixedDeltaTime, a => {
                interpolator += Time.fixedDeltaTime / transitionDuration;
                headRenderer.color = Color.Lerp(currentColor, colorClosed, interpolator);
                neckRenderer.color = Color.Lerp(currentColor, colorClosed, interpolator);
            }, (int)(transitionDuration / Time.fixedDeltaTime));
        }
        else if (state == MouthState.Idle)
        {
            headRenderer.sprite = mouthErgh;
            wantedLocalRot = Quaternion.identity;
            float interpolator = 0;
            float transitionDuration = .35f;
            Color currentColor = headRenderer.color;
            The_Helper.StartTimer(Time.fixedDeltaTime, a => {
                interpolator += Time.fixedDeltaTime / transitionDuration;
                headRenderer.color = Color.Lerp(currentColor, colorClosed, interpolator);
                neckRenderer.color = Color.Lerp(currentColor, colorClosed, interpolator);
            }, (int)(transitionDuration / Time.fixedDeltaTime));
        }
    }

    private float GetMouthSFXVolume(MouthState state)
    {
        if (state == MouthState.Open)
            return 1f;
        else if (state == MouthState.Closed)
            return .35f;
        else if (state == MouthState.Ergh)
            return .1f;

        return 1;
    }

    private void InitializeInputStream(){
        disposable = input.DataStream.Where(_ => _.player == this.player).Select(LogInput).Select(TiltToString).Subscribe(Shoot);
        input.DataStream.Where(_ => _.player != this.player).Select(GetInput).Subscribe(PrepareTurn);
    }

    private PreparationArgs LogInput(PreparationArgs data)
    {
        ToggleMouth(data.power > 0 ? MouthState.Closed : MouthState.Ergh);
        The_Helper.StartTimer(1f, a => { if (this) ToggleMouth(MouthState.Idle); });
        // GetComponent<SpriteRenderer>().color = GetComponent<SpriteRenderer>().color.Change(ColorProperty.a, 0.2f);
        Debug.Log(string.Format("Player {0} input {1} power {2}", player, data.pitch.ToString(), data.power));
        return data;
    }
    private PreparationArgs GetInput(PreparationArgs data)
    {
        // Debug.Log(string.Format("Player {0} drinking!", player));
        if (!this) return data;
        // GetComponent<SpriteRenderer>().color = 
        //    GetComponent<SpriteRenderer>().color.Change(ColorProperty.a, 1f);

        ToggleMouth(MouthState.Open);
        wantedLocalRot = Quaternion.identity;

        return data;
    }

    Potions lastConsumedPotion = Potions.Fire;
    bool potionConsumed = false;
    bool skippedTurn = false;


    private Transform GetMuzzlePoint(Pitch pitch)
    {
        switch (pitch)
        {
            case Pitch.High:
                return outputHigh;
            case Pitch.Mid:
                return outputMid;
            case Pitch.Low:
                return outputLow;
            default:
                return null;
        }
    }
    private Transform GetInstrument(Pitch pitch)
    {
        switch (pitch)
        {
            case Pitch.High:
                return instrumentHigh;
            case Pitch.Mid:
                return instrumentMid;
            case Pitch.Low:
                return instrumentLow;
            default:
                return null;
        }
    }

    private void PrepareTurn(PreparationArgs data)
    {
        if (skippedTurn)
        {
            return;
        }

        Potions consumedPotion = potionConveyorController.DrinkPotion();

        lastConsumedPotion = consumedPotion;
        potionConsumed = true;
    }


    private void Shoot(PreparationArgs data)
    {
        if (!potionConsumed)
        {
            return;
        }

        if (data.power <= 0)
        {
            skippedTurn = true;
            return;
        }
        skippedTurn = false;

        GetInstrument(data.pitch).InterpolateScale(Vector3.one * 1.35f, 0.35f, true, a =>
        {
            laserController.IMaFirinMyLazor(GetMuzzlePoint(data.pitch), data.pitch, data.power, lastConsumedPotion);
        });
        potionConveyorController.EmptySuckFill();

        // float a = GetComponent<SpriteRenderer>().color.a;

        // GetComponent<SpriteRenderer>().color = LaserLogic.GetColorOfPotion(consumedPotion).Change(ColorProperty.a, a);
    }

    internal void Damage()
    {
        healthManager.health--;

        if (healthManager.health == 0)
        {
            Debug.Log("YOU DED");

            audioSource.PlayOneShot(The_Helper.ResourceLoad<AudioClip>(
                "Death" + player), 1f);

            if (playerIsDeadCallback!=null)
			{
				playerIsDeadCallback(player);
			}

			//healthManager.health = healthManager.maxHealth;
        }
        else
        {
            audioSource.PlayOneShot(The_Helper.ResourceLoad<AudioClip>(
                "Ouch" + player), .5f);
        }
    }

    /*
	private Pitch SwapAnimation(Pitch data )
	{
		string trigger = null;
		switch( data )
		{
		case Pitch.HIGH:
			trigger = "High";
			break;
		case Pitch.MID:
			trigger = "Mid";
			break;
		case Pitch.LOW:
			trigger = "Low";
			break;
		}

		if( animator != null )
		{
			animator.SetTrigger(trigger);
		}

		return data;
	}
	*/

    Quaternion wantedLocalRot = Quaternion.identity;
    private float lerpFactor = 4;
    void Update()
    {
        headRenderer.transform.localRotation =
            Quaternion.Lerp(headRenderer.transform.localRotation, wantedLocalRot, Time.deltaTime * lerpFactor);
    }

    private PreparationArgs TiltToString(PreparationArgs data ){

        if (data.power <= 0)
        {
            wantedLocalRot = Quaternion.identity;
            return data;
        }

		switch( data.pitch )
		{
		    case Pitch.High:
                wantedLocalRot = Quaternion.Euler(Vector3.forward * 30);
			    break;
		    case Pitch.Mid:
                wantedLocalRot = Quaternion.Euler(Vector3.zero);
                break;
		    case Pitch.Low:
                wantedLocalRot = Quaternion.Euler(Vector3.forward * - 30);
                break;
		}

		return data;
	}

	private void OnDestroy()
	{
		disposable.Dispose();
	}

	private class TestInputDevice:MicrophoneInput{

		public IObservable<PreparationArgs> DataStream{get; private set;}

		private Player player ;
		private int playerId =-1;

		private IDisposable pitchListener;

		private KeyCode[][] keysForTest = new KeyCode[][]{
			new KeyCode[]{ KeyCode.Q, KeyCode.A, KeyCode.Z},
			new KeyCode[]{ KeyCode.I, KeyCode.K, KeyCode.M},
		};

		public TestInputDevice( Player player )
		{
			this.player = player;
			this.playerId = (Player.Left == player)?0:1;

			DataStream = new Subject<PreparationArgs>();
		}

		public void Reset()
		{
				pitchListener =
				Observable.EveryUpdate().Select( _ => Input.GetKey(keysForTest[playerId][0]) ).DistinctUntilChanged().Where(_=>_).Select(_=>Pitch.High).Merge( 
				Observable.EveryUpdate().Select( _ => Input.GetKey(keysForTest[playerId][1]) ).DistinctUntilChanged().Where(_=>_).Select(_=>Pitch.Mid).Merge(
				Observable.EveryUpdate().Select( _ => Input.GetKey(keysForTest[playerId][2]) ).DistinctUntilChanged().Where(_=>_).Select(_=>Pitch.Low).Merge(
					))).Subscribe( OnPitch );
		}

		private void OnPitch( Pitch pitch )
		{
			pitchListener.Dispose();

			float timer = 0;

			IDisposable selfDisposable = null;

			selfDisposable = Observable.EveryUpdate().Subscribe( _ => { 
				timer += Time.deltaTime;

				if( !Input.GetKey(keysForTest[playerId][0]) && !Input.GetKey(keysForTest[playerId][1]) &&  !Input.GetKey(keysForTest[playerId][2])  ){
					((Subject<PreparationArgs>)DataStream).OnNext(  new PreparationArgs( player,  timer, pitch ) );

					selfDisposable.Dispose();
				}	

			});

		}

	}

}

public enum MouthState { Idle, Open, Closed, Ergh }